"""
Manage thread local states
"""

import random
import threading
import weakref
from contextlib import contextmanager
from typing import Any, Optional, Tuple, Union

import jax
import jax.numpy as jnp
import jax.tree_util

KeyArray = Union[Any, jnp.ndarray]


class PaxThreadingLocalState(threading.local):
    """Manage all thread local states used by PAX"""

    _mutable_module_ref_list: Tuple[weakref.ReferenceType, ...]
    _mutable_module_level: int
    _rng: Optional[random.Random]

    def __init__(self):
        super().__init__()
        self._mutable_module_ref_list = ()
        self._mutable_module_level = _jax_cur_level()
        self._rng = random.Random(42)

    def add_mutable_module(self, module):
        """add `module` to mutable list"""
        self._mutable_module_ref_list = (
            weakref.ref(module),
            *self._mutable_module_ref_list,
        )

    def is_mutable(self, module):
        """Is `module` mutable?"""

        # cannot modify a module whose level of abstraction
        # is lower than the current level
        if self._mutable_module_level < _jax_cur_level():
            return False

        for ref in self._mutable_module_ref_list:
            if module is ref():
                return True

        return False

    @contextmanager
    def allow_mutation(self, modules):
        """A context manager that turns on mutability."""

        if not isinstance(modules, (tuple, list)):
            modules = (modules,)
        modules = tuple(weakref.ref(mod) for mod in modules)

        prev = self._mutable_module_ref_list
        prev_abstraction_level = self._mutable_module_level
        try:
            self._mutable_module_ref_list = modules
            self._mutable_module_level = _jax_cur_level()
            yield
        finally:
            self._mutable_module_ref_list = prev
            self._mutable_module_level = prev_abstraction_level

    def seed_rng_key(self, seed: int) -> None:
        """Set ``self._rng = random.Random(seed)``.

        Arguments:
            seed: an integer seed.
        """
        assert isinstance(seed, int)
        self._rng = random.Random(seed)

    def next_rng_key(self) -> KeyArray:
        """Return a random rng key. Renew the global random state."""
        seed = self._rng.randint(1, 999999999)
        return jax.random.PRNGKey(seed)

    def get_rng_state(self):
        """Return internal random states."""
        return self._rng.getstate()

    def set_rng_state(self, state):
        """Set internal random states."""
        self._rng.setstate(state)


def _jax_cur_level():
    """
    Return the level of current jax trace.

    If it is an eval_trace, return -1.
    """
    trace = jax.core.thread_local_state.trace_state.trace_stack.stack[-1]
    if trace.trace_type == jax.core.EvalTrace:
        return -1
    else:
        return trace.level


PAX_STATE = PaxThreadingLocalState()
add_mutable_module = PAX_STATE.add_mutable_module
allow_mutation = PAX_STATE.allow_mutation
get_rng_state = PAX_STATE.get_rng_state
is_mutable = PAX_STATE.is_mutable
next_rng_key = PAX_STATE.next_rng_key
seed_rng_key = PAX_STATE.seed_rng_key
set_rng_state = PAX_STATE.set_rng_state



"""Transform a module to a new one."""
from typing import Any, TypeVar

import jax

from .module import Module, parameters_method, update_pytree

TreeDef = Any

T = TypeVar("T", bound=Module)
K = TypeVar("K", bound=Module)
O = TypeVar("O", bound=Module)


def enable_train_mode(mod: T) -> T:
    """Return a module in training mode."""
    return mod.train()


def enable_eval_mode(mod: T) -> T:
    """Return a module in evaluation mode."""
    return mod.eval()


def freeze_parameters(mod: T) -> T:
    """Return a copy module with all trainable parameters are converted to non-trainable states."""

    def _freeze_apply_fn(mod: T) -> T:
        return mod.replace_method(parameters=parameters_method())

    return mod.apply(_freeze_apply_fn)


def unfreeze_parameters(mod: T, *, origin: T) -> T:
    """Return a copy module with all trainable parameters are converted to non-trainable states."""
    tree_def = jax.tree_structure(origin)
    leaves = jax.tree_leaves(mod)
    return jax.tree_unflatten(tree_def, leaves)


def select_parameters(mod: T) -> T:
    """Select `PARAMETER` leaves only."""
    return mod.parameters()


def update_parameters(mod: T, *, params: T) -> T:
    """Return a module that uses trainable parameters in `params`."""
    return update_pytree(mod, other=params.parameters())



"""Safeguards to prevent potential bugs."""

import inspect
from typing import Iterable, List, Type, TypeVar

import jax
import jax.numpy as jnp
import numpy as np

from .base import BaseModule
from .threading_local import allow_mutation, is_mutable

T = TypeVar("T")


class SafeBaseModuleMetaclass(type):
    """Metaclass for `SafeBaseModule`."""

    def __call__(cls: Type[T], *args, **kwargs) -> T:
        module = cls.__new__(cls, *args, **kwargs)  # type: ignore

        with allow_mutation(module):
            cls.__init__(module, *args, **kwargs)
            module.find_and_register_pytree_attributes()

        # scan module after initialization for potential bugs
        if hasattr(module, "__slots__"):
            raise ValueError("`__slots__` is not supported by PAX modules.")
        module._assert_not_shared_module()
        module._assert_not_shared_weight()
        module._scan_fields(module._class_fields())
        return module


class SafeBaseModule(BaseModule, metaclass=SafeBaseModuleMetaclass):
    """Adding safe guards to BaseModule to prevent bugs."""

    def _class_fields(self):
        for name, value in inspect.getmembers(self):
            if name.startswith("__") or inspect.ismethod(value):
                continue

            if name in self.__dict__:
                continue

            if find_descriptor(self.__class__, name) is not None:
                # ignore descriptors
                continue

            yield name

    def _assert_mutability(self):
        if not is_mutable(self):
            raise ValueError(
                "Cannot modify a module in immutable mode.\n"
                "Please do this computation inside a function decorated by `pax.pure`."
            )

    def _assert_not_shared_module(self):
        """Shared module is not allowed."""
        shared_module = _find_shared_module(self)
        if shared_module is not None:
            raise ValueError(
                f"The module `{shared_module}` is shared between two nodes of the pytree.\n"
                f"This is not allowed to prevent potential silence bugs."
            )

    def _assert_not_shared_weight(self):
        """Shared weight is not allowed."""
        leaves = jax.tree_leaves(self)
        leaf_ids = set()
        for leaf in leaves:
            if id(leaf) in leaf_ids:
                raise ValueError(
                    f"Detected a shared ndarray. This is not allowed.\n"
                    f"Shape={leaf.shape}\n"
                    f"Dtype={leaf.dtype}\n"
                    f"Value={leaf}",
                )
            leaf_ids.add(id(leaf))

    def _scan_fields(self, fields: Iterable[str]):
        """Scan fields for *potential* bugs."""

        for name in fields:
            if name in self.pytree_attributes:
                continue

            value = getattr(self, name)
            is_mod = lambda x: isinstance(x, BaseModule)
            is_ndarray = lambda x: isinstance(x, (jnp.ndarray, np.ndarray))
            mods, _ = jax.tree_flatten(value, is_leaf=is_mod)
            leaves = jax.tree_leaves(value)
            has_mods = any(map(is_mod, mods))
            has_arrays = any(map(is_ndarray, mods))

            if has_mods:
                raise ValueError(
                    f"\n"
                    f"Unregistered field `{self}.{name}`:\n"
                    f"    value={value}\n"
                    f"contains a module leaf.\n"
                )

            if has_arrays:
                raise ValueError(
                    f"\n"
                    f"Unregistered field `{self}.{name}`:\n"
                    f"    value={value}\n"
                    f"contains a ndarray leaf.\n"
                )


def _find_shared_module(module: BaseModule):
    """Find shared module.

    - Return the first module that is shared by two nodes of the pytree.
    - Return `None` if there is no shared module.
    """

    def _get_all_modules(mod: BaseModule, lst: List):
        lst.append(mod)
        is_mod = lambda x: isinstance(x, BaseModule) and x is not mod
        submodules, _ = jax.tree_flatten(mod, is_leaf=is_mod)
        submodules = (m for m in submodules if is_mod(m))
        for m in submodules:
            _get_all_modules(m, lst)

    mods = []
    _get_all_modules(module, mods)
    module_ids = set()
    for m in mods:
        if id(m) in module_ids:
            return m
        module_ids.add(id(m))

    return None


# source: https://stackoverflow.com/a/21963090
def find_descriptor(cls, attrname):
    """Find the descriptor of an attribute."""

    def hasspecialmethod(obj, name):
        return any(name in klass.__dict__ for klass in type(obj).__mro__)

    for klass in cls.__mro__:
        if attrname in klass.__dict__:
            descriptor = klass.__dict__[attrname]
            if not hasspecialmethod(descriptor, "__get__"):
                return None
            return descriptor
    return None



"""PAX mechanisms to make PAX functions pure."""

import functools
from types import MethodType
from typing import Any, Callable, Tuple, TypeVar

import jax

from .base import BaseModule
from .threading_local import allow_mutation

T = TypeVar("T")
O = TypeVar("O")


def pure(func: Callable):
    """Make a function pure by copying the inputs.

    Any modification on the copy will not affect the original inputs.

    **Note**: only functions that are wrapped by `pax.pure` are allowed to modify PAX's Modules.

    Example:

    >>> f = pax.Linear(3,3)
    >>> f.a_list = []
    Traceback (most recent call last):
      ...
    ValueError: Cannot modify a module in immutable mode.
    Please do this computation inside a function decorated by `pax.pure`.
    >>>
    >>> @pax.pure
    ... def add_list(m):
    ...     m.a_list = []
    ...     return m
    ...
    >>> f = add_list(f)
    >>> print(f.a_list)
    []

    Arguments:
        func: A function.

    Returns:
        A pure function.
    """

    @functools.wraps(func)
    def wrapper(*args, **kwargs):
        for m in _get_modules((func, args, kwargs)):
            m.scan_bugs()

        # support calling method
        if isinstance(func, MethodType):
            args = (func.__self__, *args)
            unbound_func = func.__func__
        # or calling a module
        elif isinstance(func, BaseModule) and callable(func):
            args = (func, *args)
            unbound_func = func.__call__.__func__
        elif callable(func):
            unbound_func = func
        else:
            raise ValueError("Not supported")

        args, kwargs = _copy((args, kwargs))
        modules = get_all_submodules((args, kwargs))
        with allow_mutation(modules):
            out = unbound_func(*args, **kwargs)

            for m in modules:
                m.find_and_register_pytree_attributes()
                m.scan_bugs()
        return out

    return wrapper


@pure
def purecall(module: Callable[..., O], *args, **kwargs) -> Tuple[Any, O]:
    """Call a module and return the updated module.

    A shortcut for `pax.pure(lambda f, x: [f, f(x)])`.
    """
    assert isinstance(module, BaseModule)
    assert callable(module)
    return module, module(*args, **kwargs)


def _get_modules(tree):
    "Return a list of modules in the pytree `tree`."
    modules = jax.tree_flatten(tree, is_leaf=lambda x: isinstance(x, BaseModule))[0]
    modules = [m for m in modules if isinstance(m, BaseModule)]
    return modules


def get_all_submodules(value):
    submods = _get_modules(value)
    out = list(submods)
    for mod in submods:
        out.extend(get_all_submodules(mod.submodules()))
    return out


def _copy(value: T) -> T:
    leaves, treedef = jax.tree_flatten(value)
    return jax.tree_unflatten(treedef, leaves)



"""PAX mechanisms to make PAX method pure."""

from functools import partial
from types import MethodType
from typing import Callable, Tuple, TypeVar

from .base import BaseModule
from .pure import pure

O = TypeVar("O")
T = TypeVar("T", bound=BaseModule)


def module_and_value(module_or_method: Callable[..., O]) -> Callable[..., Tuple[T, O]]:
    """Return a pure function that executes a module's method.

    This pure function also returns the updated input module in the output.

    Example:

    >>> net = pax.Linear(1, 1)
    >>> x = jnp.ones((32, 1))
    >>> net, y = pax.module_and_value(net)(x)  # note: `net` is also returned.


    Arguments:
        module_or_method: Either a PAX module or a method of a PAX module.

    Returns:
        A pure function.
    """
    is_bound_method = True
    if isinstance(module_or_method, MethodType):  # a method
        mod = module_or_method.__self__
        func = module_or_method.__func__
    elif isinstance(module_or_method, BaseModule):  # a module
        mod = module_or_method
        assert hasattr(mod, "__call__"), "Expecting a callable module."
        func = module_or_method.__call__.__func__
    elif callable(module_or_method):
        is_bound_method = False
        func = module_or_method
    else:
        raise ValueError("Expecting a module or a module's method.")

    @pure
    def _run(mod, *args, **kwargs):
        assert isinstance(mod, BaseModule), "Expecting a PAX module."
        out = func(mod, *args, **kwargs)
        return mod, out

    if is_bound_method:
        return partial(_run, mod)
    else:
        return _run



"""PAX BaseModule."""

# Note: This file is originated from
# https://raw.githubusercontent.com/cgarciae/treex/32e4cce5ca0cc991cda8076903853621d0aa4ab9/treex/module.py
# which is under MIT License.

from typing import Any, List, Mapping, Optional, Tuple, TypeVar

import jax
import jax.numpy as jnp
import jax.tree_util
import numpy as np

T = TypeVar("T", bound="BaseModule")
M = TypeVar("M")


class BaseModule:
    """BaseModule manages all information related to the pytree.

    There are two important methods:

    - ``tree_flatten`` converts a module to ``(leaves, treedef)``
    - ``tree_unflatten`` restores the module.

    BaseModule maintains a ``pytree_attributes`` tuple that lists all subtree attribute names.
    """

    _pytree_attributes: Tuple[str, ...] = ()
    _mixed_pytree_attributes: Optional[Tuple[str, ...]] = None

    @property
    def pytree_attributes(self):
        if self._mixed_pytree_attributes is not None:
            return self._pytree_attributes + self._mixed_pytree_attributes
        else:
            return self._pytree_attributes

    def find_and_register_pytree_attributes(self: T):
        """Find and register ndarrays and submodules."""
        is_mod_or_node = lambda x: isinstance(x, (BaseModule, EmptyNode))
        is_pytree = lambda x: isinstance(x, pytree_cls)

        pytree_attributes = []
        mixed_pytree_attributes = []
        for name, value in self.__dict__.items():
            leaves, _ = jax.tree_flatten(value, is_leaf=is_mod_or_node)
            pytree_cls = (jnp.ndarray, np.ndarray, BaseModule, EmptyNode)
            any_pytree = any(map(is_pytree, leaves))
            all_pytree = all(map(is_pytree, leaves))
            if any_pytree and all_pytree:
                pytree_attributes.append(name)
            elif any_pytree:
                mixed_pytree_attributes.append(name)
        self._pytree_attributes = tuple(pytree_attributes)
        if len(mixed_pytree_attributes) > 0:
            self._mixed_pytree_attributes = tuple(mixed_pytree_attributes)
        else:
            self._mixed_pytree_attributes = None

    def tree_flatten(self) -> Tuple[List[jnp.ndarray], Mapping[str, Any]]:
        """Convert a module to ``(children, treedef)``."""
        aux = dict(self.__dict__)
        children = [aux.pop(name) for name in self._pytree_attributes]
        if self._mixed_pytree_attributes is not None:
            is_module = lambda x: isinstance(x, BaseModule)
            array_mod_cls = (jnp.ndarray, np.ndarray, BaseModule)
            is_array_mod = lambda x: isinstance(x, array_mod_cls)
            for name in self._mixed_pytree_attributes:
                value = aux.pop(name)
                leaves, treedef = jax.tree_flatten(value, is_leaf=is_module)
                leaves = (v if is_array_mod(v) else ValueNode(v) for v in leaves)
                value = jax.tree_unflatten(treedef, leaves)
                children.append(value)
        return children, aux

    @classmethod
    def tree_unflatten(cls, aux, children):
        """Recreate a module from its ``(children, treedef)``."""
        module = object.__new__(cls)
        module_dict = module.__dict__
        module_dict.update(aux)
        module_dict.update(zip(module._pytree_attributes, children))
        if module._mixed_pytree_attributes is not None:
            L = len(module._pytree_attributes)
            is_leaf = lambda x: isinstance(x, (ValueNode, BaseModule))
            unwrap = lambda x: x.value if isinstance(x, ValueNode) else x
            for name, value in zip(module._mixed_pytree_attributes, children[L:]):
                module_dict[name] = jax.tree_map(unwrap, value, is_leaf=is_leaf)
        return module

    def __init_subclass__(cls):
        """Any subclass of ``Module`` is also registered as pytree."""
        jax.tree_util.register_pytree_node_class(cls)

    def __eq__(self, o: object) -> bool:
        """Compare two modules."""
        if id(self) == id(o):
            return True

        if type(self) is not type(o):
            return False

        self_leaves, self_treedef = jax.tree_flatten(self)
        o_leaves, o_treedef = jax.tree_flatten(o)

        if len(self_leaves) != len(o_leaves):
            return False

        if self_treedef != o_treedef:
            return False

        leaves_equal = jax.tree_map(lambda a, b: a is b, self_leaves, o_leaves)
        return all(leaves_equal)

    def __hash__(self) -> int:
        leaves, treedef = jax.tree_flatten(self)
        leaves = jax.tree_map(lambda x: (x.shape, x.dtype), leaves)
        return hash((tuple(leaves), treedef))


#  Note: this class is inspired by treex's `Nothing` class.
@jax.tree_util.register_pytree_node_class
class EmptyNode:
    """Mark an uninitialized or deleted pytree node."""

    def tree_flatten(self):
        """Flatten empty node."""
        return (), None

    @classmethod
    def tree_unflatten(cls, aux, children):
        """Unflatten empty node."""
        del aux, children
        return EmptyNode()

    def __repr__(self) -> str:
        return "EmptyNode"

    def __eq__(self, o: object) -> bool:
        if isinstance(o, EmptyNode):
            return True
        return False


@jax.tree_util.register_pytree_node_class
class ValueNode:
    """We use this class to store a value in treedef."""

    def __init__(self, value):
        super().__init__()
        self.value = value

    def tree_flatten(self):
        return (), self.value

    @classmethod
    def tree_unflatten(cls, value, children):
        return ValueNode(value)

    def __repr__(self) -> str:
        return f"ValueNode({self.value})"



from contextlib import contextmanager

from .module import Module
from .pure import get_all_submodules
from .threading_local import allow_mutation


@contextmanager
def mutable(module: Module):
    """A context manager that allows a copy module to be mutable inside the context.

    >>> net = pax.Linear(1, 2)
    >>> with pax.experimental.mutable(net) as net:
    ...     net.bias = jnp.array(0.)
    >>> assert net.bias.item() == 0.
    """

    copy = module.copy()
    all_submodules = get_all_submodules(copy)

    with allow_mutation(all_submodules):
        try:
            yield copy
        finally:
            copy.find_and_register_pytree_attributes()
            copy.scan_bugs()



"""PAX module."""

from typing import Any, Callable, Dict, List, Optional, Tuple, TypeVar, Union

import jax
import jax.numpy as jnp
import jax.tree_util
import numpy as np

from .base import EmptyNode, ValueNode
from .module_and_value import module_and_value
from .pure import pure
from .safe_module import SafeBaseModule
from .threading_local import allow_mutation

T = TypeVar("T", bound="Module")
M = TypeVar("M")
TreeDef = Any


def parameters_method(*trainable_attributes, submodules=True):
    """Return a `parameters` method.

    >>> class Linear(pax.Module):
    ...     parameters = pax.parameters_method("weight")
    ...     def __init__(self):
    ...         self.weight = jnp.array(1.0)
    >>> fc = Linear()
    >>> fc == fc.parameters()
    True

    Arguments:
        trainable_atributes: a list of trainable attribute names.
        submodules: include submodules if true.

    Returns:
        A method that returns a module with only trainable weights.
    """

    names = []
    for name in trainable_attributes:
        names.extend(name.split())

    def _parameters(self: T) -> T:
        for name in names:
            assert hasattr(self, name), f"Expecting an attribute with name `{name}`."

        is_submodule = lambda x: x is not self and isinstance(x, SafeBaseModule)
        leaves, treedef = jax.tree_flatten(self, is_leaf=is_submodule)
        leaves = (
            leaf.parameters() if submodules and is_submodule(leaf) else EmptyNode()
            for leaf in leaves
        )
        mod = jax.tree_unflatten(treedef, leaves)
        values = {name: getattr(self, name) for name in names}
        mod = mod.replace(**values)
        return mod

    return _parameters


class Module(SafeBaseModule):
    """The base class for all PAX modules.

    Example:

    >>> class Counter(pax.Module):
    ...     def __init__(self):
    ...         super().__init__()
    ...         self.count = jnp.array(0)
    ...
    ...     def step(self, x):
    ...         self.count += 1
    """

    _name: Optional[str]
    _training: bool

    def __new__(cls, *args, **kwargs):
        """Creata a new module.

        Add `_name` and `_training` to __dict__ to make sure that
        the module works correctly even if the __init__ method is
        not called.
        """
        del args, kwargs
        obj = super().__new__(cls)
        obj.__dict__["_name"] = None
        obj.__dict__["_training"] = True
        return obj

    def __init__(self, *, training: bool = True, name: Optional[str] = None):
        """Initialize module."""
        super().__init__()
        self._name = name
        self._training = training

    @property
    def name(self):
        return self._name

    @property
    def training(self) -> bool:
        """Return `True` if a module is in training mode.

        >>> net = pax.Linear(1, 1)
        >>> net.training
        True
        >>> net = net.eval()
        >>> net.training
        False
        """
        return self._training

    def parameters(self):
        """Return a new module with trainable weights only."""
        return parameters_method(submodules=True)(self)

    def copy(self: T) -> T:
        """Return a copy of the current module."""
        leaves, treedef = jax.tree_flatten(self)
        return jax.tree_unflatten(treedef, leaves)

    def train(self: T) -> T:
        """Return a module in training mode."""
        return self.apply(lambda mod: mod.replace(_training=True))

    def eval(self: T) -> T:
        """Return a module in evaluation mode."""
        return self.apply(lambda mod: mod.replace(_training=False))

    def update_parameters(self: T, params: T) -> T:
        """Return a new module with updated parameters."""
        return update_pytree(self, other=params.parameters())

    def replace_method(self: T, **methods) -> T:
        cls = self.__class__
        cls_name = cls.__name__
        cls = type(cls_name, (cls,), methods)
        obj = object.__new__(cls)
        obj.__dict__.update(self.__dict__)
        return obj

    def replace(self: T, **kwargs) -> T:
        """Return a new module with some attributes replaced.

        >>> net = pax.Linear(2, 2)
        >>> net = net.replace(bias=jnp.zeros((2,)))
        """

        mod = self.copy()
        with allow_mutation(mod):
            for name, value in kwargs.items():
                assert hasattr(mod, name)
                setattr(mod, name, value)
            mod.find_and_register_pytree_attributes()

        mod.scan_bugs()
        return mod

    # inspired by patrick-kidger/equinox `tree_at`
    def replace_node(self: T, node: jnp.ndarray, value: jnp.ndarray) -> T:
        """Replace a node of the pytree by a new value.

        Example:

        >>> mod = pax.Sequential(
        ...     pax.Linear(2,2),
        ...     jax.nn.relu
        ... )
        >>> mod = mod.replace_node(mod[0].weight, jnp.zeros((2, 3)))
        >>> print(mod[0].weight.shape)
        (2, 3)
        """
        leaves, tree_def = jax.tree_flatten(self, is_leaf=lambda x: x is node)
        count = sum(1 if x is node else 0 for x in leaves)

        if count != 1:
            raise ValueError(f"The node `{node}` appears {count} times in the module.")

        # replace `node` by value
        new_leaves = [value if v is node else v for v in leaves]
        mod: T = jax.tree_unflatten(tree_def, new_leaves)
        mod.scan_bugs()
        return mod

    def set_attribute(self: T, name, value) -> T:
        """Create a new module and set attribute."""
        module = self.copy()
        with allow_mutation(module):
            setattr(module, name, value)
            module.find_and_register_pytree_attributes()

        return module

    def summary(self, return_list: bool = False) -> Union[str, List[str]]:
        """Summarize a module as a tree of its submodules.

        Arguments:
            return_list: return a list of lines instead of a joined string.

        >>> net = pax.Sequential(pax.Linear(2, 3), jax.nn.relu, pax.Linear(3, 4))
        >>> print(net.summary())
        Sequential
        ├── Linear(in_dim=2, out_dim=3, with_bias=True)
        ├── x => relu(x)
        └── Linear(in_dim=3, out_dim=4, with_bias=True)
        """

        output = [self.__repr__()]
        if output[0] is None:
            raise ValueError(
                f"The `{self.__class__}.__repr__` method returns a `None` value."
            )
        submodules: List[Module] = self.submodules()

        def indent(lines: List[str], start_string) -> List[str]:
            return [start_string + l for l in lines]

        for i, module in enumerate(submodules):
            lines = module.summary(return_list=True)
            if i + 1 < len(submodules):  # middle submodules
                indented_lines = indent(lines[:1], "├── ") + indent(lines[1:], "│   ")
            else:  # last submodule
                indented_lines = indent(lines[:1], "└── ") + indent(lines[1:], "    ")
            output.extend(indented_lines)

        if return_list:
            return output
        else:
            return "\n".join(output)

    def map(self: T, func, *mods) -> T:
        mods = [mods.state_dict() for mods in mods]
        return self.load_state_dict(jax.tree_map(func, self.state_dict(), *mods))

    def apply(self: T, apply_fn) -> T:
        """Apply a function to all submodules.

        >>> def print_param_count(mod):
        ...     count = sum(jax.tree_leaves(jax.tree_map(jnp.size, mod)))
        ...     print(f"{count}\t{mod}")
        ...     return mod
        ...
        >>> net = pax.Sequential(pax.Linear(1, 1), jax.nn.relu)
        >>> net = net.apply(print_param_count)
        2 Linear(in_dim=1, out_dim=1, with_bias=True)
        0 Lambda(relu)
        2 Sequential

        Arguments:
            apply_fn: a function which inputs a module and outputs a transformed module.
        """

        def rec_fn(mod_or_ndarray):
            if isinstance(mod_or_ndarray, Module):
                return mod_or_ndarray.apply(apply_fn)
            else:
                return mod_or_ndarray

        submodules: List[Module] = self.submodules()
        new_self = jax.tree_map(
            rec_fn,
            self,
            is_leaf=lambda x: isinstance(x, Module) and (x in submodules),
        )

        # tree_map already created a copy of self,
        # hence `apply_fn` is guaranteed to have no side effects.
        return apply_fn(new_self)

    def apply_submodules(self: M, func: Callable[..., M]) -> M:
        """Apply a function to all submodules, recursively."""
        module = self.copy()
        submod_fn = lambda x: isinstance(x, Module) and x is not module
        leaves, treedef = jax.tree_flatten(module, is_leaf=submod_fn)
        new_leaves = []
        for value in leaves:
            if isinstance(value, Module):
                new_leaves.append(value.apply(func))
            else:
                new_leaves.append(value)
        return jax.tree_unflatten(treedef, new_leaves)

    def submodules(self) -> List[T]:
        """Return a list of submodules."""
        submod_fn = lambda x: isinstance(x, Module) and x is not self
        leaves, _ = jax.tree_flatten(self, is_leaf=submod_fn)
        return [leaf for leaf in leaves if submod_fn(leaf)]

    def scan_bugs(self: T) -> T:
        """Scan the module for potential bugs."""

        # scan for shared module/weight.
        self._assert_not_shared_module()
        self._assert_not_shared_weight()

        def _scan_field_fn(mod: T) -> T:
            assert isinstance(mod, Module)
            # pylint: disable=protected-access
            mod._scan_fields(mod._class_fields())
            return mod

        self.apply(_scan_field_fn)
        return self

    def state_dict(self) -> Dict[str, Any]:
        """Return module's state dictionary."""
        return save_weights_to_dict(self)

    def load_state_dict(self: T, state_dict: Dict[str, Any]) -> T:
        """Return a new module from the state dictionary."""
        return load_weights_from_dict(self, state_dict)

    def __setattr__(self, name: str, value: Any) -> None:
        self._assert_mutability()
        super().__setattr__(name, value)
        # prevent infinite loop
        if name not in ("_pytree_attributes", "_mixed_pytree_attributes"):
            self.find_and_register_pytree_attributes()

    def __delattr__(self, name: str) -> None:
        self._assert_mutability()
        if name in self.pytree_attributes:
            raise ValueError(
                "Cannot delete a pytree attribute.\n"
                "This is to avoid potential bugs related to the order of pytree attributes.\n"
            )
        super().__delattr__(name)

    def __mod__(self: T, args: Union[Any, Tuple]) -> Tuple[T, Any]:
        """An alternative to `pax.module_and_value`.

        >>> bn = pax.BatchNorm1D(3)
        >>> x = jnp.ones((5, 8, 3))
        >>> bn, y = bn % x
        >>> bn
        BatchNorm1D(num_channels=3, ...)
        """
        assert callable(self)

        if isinstance(args, tuple):
            return module_and_value(self)(*args)
        else:
            return module_and_value(self)(args)

    def __or__(self: T, other: T) -> T:
        """Merge two modules.

        >>> a = pax.Linear(2, 2)
        >>> b = pax.Linear(2, 2)
        >>> c = a | b
        >>> d = b | a
        >>> c == d
        False
        """
        return update_pytree(self, other=other)

    def __invert__(self: T) -> T:
        return self.parameters()

    def _repr(self, info: Optional[Dict[str, Any]] = None) -> str:
        name = f"({self.name}) " if self.name is not None else ""
        cls_name = self.__class__.__qualname__
        if info is None:
            return f"{name}{cls_name}"
        else:
            lst_info = [f"{k}={v}" for (k, v) in info.items() if v is not None]
            str_info = ", ".join(lst_info)
            return f"{name}{cls_name}({str_info})"

    def __repr__(self) -> str:
        return self._repr()


def update_pytree(mod: T, *, other: T) -> T:
    """Use non-EmptyNode leaves from other."""

    def _select_fn(leaf_x, leaf_y):
        if isinstance(leaf_y, EmptyNode):
            return leaf_x
        else:
            return leaf_y

    is_empty = lambda x: isinstance(x, EmptyNode)
    new_mod = jax.tree_map(_select_fn, mod, other, is_leaf=is_empty)
    return new_mod


def save_weights_to_dict(module: Module) -> Dict[str, Any]:
    """Save module weights to a dictionary.

    >>> net = pax.Sequential(pax.Linear(1, 2), jax.nn.relu, pax.Linear(2, 3))
    >>> weights = pax.experimental.save_weights_to_dict(net)
    >>> weights
    {'modules': ({'weight': ..., 'bias': ...}, {}, {'weight':..., 'bias': ...})}
    """

    def _dict_or_array(v):
        if isinstance(v, Module):
            return save_weights_to_dict(v)
        elif isinstance(v, (jnp.ndarray, np.ndarray)):
            return v
        else:
            return None

    out = {}
    for name in module.pytree_attributes:
        value = getattr(module, name)
        out[name] = jax.tree_map(
            _dict_or_array,
            value,
            is_leaf=lambda x: isinstance(x, Module),
        )

    return out


@pure
def load_weights_from_dict(module: T, state_dict: Dict[str, Any]) -> T:
    """Load module weights from a dictionary.

    >>> a = pax.Sequential(pax.Linear(1, 2), jax.nn.relu, pax.Linear(2, 3))
    >>> weights = pax.experimental.save_weights_to_dict(a)
    >>> b = pax.Sequential(pax.Linear(1, 2), jax.nn.relu, pax.Linear(2, 3))
    >>> b = pax.experimental.load_weights_from_dict(b, weights)
    >>> assert a == b
    """

    def _module_or_array(m, s):
        if isinstance(m, Module):
            return load_weights_from_dict(m, s)
        elif isinstance(m, (jnp.ndarray, np.ndarray)):
            return s
        elif s is None:
            return m
        else:
            raise ValueError("Impossible")

    out = module
    for name in module.pytree_attributes:
        value = getattr(module, name)
        value = jax.tree_map(
            _module_or_array,
            value,
            state_dict[name],
            is_leaf=lambda x: isinstance(x, Module),
        )
        setattr(out, name, value)

    return out



"""EMA module."""

from typing import Any, Optional

import jax
import jax.numpy as jnp

from ..core import StateModule


def _has_integer_leaves(x):
    """check if there is any interger/bool leaves"""
    leaves = jax.tree_leaves(x)
    return not all(jnp.issubdtype(leaf, jnp.floating) for leaf in leaves)


class EMA(StateModule):
    """Exponential Moving Average (EMA) Module"""

    averages: Any
    decay_rate: float
    debias: Optional[jnp.ndarray]
    allow_int: bool

    def __init__(
        self,
        initial_value,
        decay_rate: float,
        debias: bool = False,
        allow_int: bool = False,
    ):
        """Create a new EMA module.

        If allow_int=True, integer leaves are updated to
        the newest values instead of averaging.

        Arguments:
            initial_value: the initial value.
            decay_rate: the decay rate.
            debias: ignore the initial value to avoid biased estimates.
            allow_int: allow integer values.
        """
        if not allow_int:
            if _has_integer_leaves(initial_value):
                raise ValueError(
                    "There are integer arrays in the initial value.\n"
                    "Use `allow_int=True` to allow this."
                )

        super().__init__()
        self.averages = initial_value
        self.decay_rate = decay_rate
        self.allow_int = allow_int
        if debias:
            # avoid integer ndarray for `jax.grad` convenience,
            # e.g., no need to pass `allow_int=True` to `jax.grad`.
            self.debias = jnp.array(0.0)
        else:
            self.debias = None

    def __call__(self, xs):
        """Return the ema of `xs`. Also, update internal states."""
        if not self.allow_int:
            if _has_integer_leaves(xs):
                raise ValueError(
                    "There are integer arrays in the new value.\n"
                    "Use `allow_int=True` to allow this."
                )

        if self.training:
            if self.debias is not None:
                cond = self.debias > 0
                debias_func = lambda a, x: jnp.where(cond, a, x)
                self.debias = jnp.array(1.0)
            else:
                debias_func = lambda a, _: a

            def update_fn(a, x):
                if jnp.issubdtype(a, jnp.floating):
                    a = debias_func(a, x)
                    return a * self.decay_rate + x * (1 - self.decay_rate)
                else:
                    return x

            self.averages = jax.tree_map(update_fn, self.averages, xs)

        return self.averages



"""PAX package."""

from pax import experimental, nets, utils
from pax._src.core import (
    EmptyNode,
    Module,
    ParameterModule,
    StateModule,
    apply_mp_policy,
    assert_structure_equal,
    enable_eval_mode,
    enable_train_mode,
    freeze_parameters,
    module_and_value,
    parameters_method,
    pure,
    purecall,
    select_parameters,
    unfreeze_parameters,
    unwrap_mp_policy,
    update_parameters,
)
from pax._src.core.rng import next_rng_key, seed_rng_key
from pax._src.nn import (
    EMA,
    GRU,
    LSTM,
    BatchNorm1D,
    BatchNorm2D,
    Conv1D,
    Conv1DTranspose,
    Conv2D,
    Conv2DTranspose,
    Dropout,
    Embed,
    GroupNorm,
    GRUState,
    Identity,
    Lambda,
    LayerNorm,
    Linear,
    LSTMState,
    MultiHeadAttention,
    RngSeq,
    Sequential,
    VanillaRNN,
    VanillaRNNState,
    avg_pool,
    max_pool,
)
from pax._src.nn.dropout import dropout
from pax._src.utils import build_update_fn, grad, scan, value_and_grad

__version__ = "0.5.8"

__all__ = (
    "apply_mp_policy",
    "assert_structure_equal",
    "avg_pool",
    "BatchNorm1D",
    "BatchNorm2D",
    "build_update_fn",
    "Conv1D",
    "Conv1DTranspose",
    "Conv2D",
    "Conv2DTranspose",
    "dropout",
    "Dropout",
    "EMA",
    "Embed",
    "EmptyNode",
    "enable_eval_mode",
    "enable_train_mode",
    "experimental",
    "freeze_parameters",
    "grad",
    "GroupNorm",
    "GRU",
    "GRUState",
    "Identity",
    "Lambda",
    "LayerNorm",
    "Linear",
    "LSTM",
    "LSTMState",
    "max_pool",
    "module_and_value",
    "Module",
    "MultiHeadAttention",
    "nets",
    "next_rng_key",
    "ParameterModule",
    "parameters_method",
    "pure",
    "purecall",
    "RngSeq",
    "scan",
    "seed_rng_key",
    "select_parameters",
    "Sequential",
    "StateModule",
    "unfreeze_parameters",
    "unwrap_mp_policy",
    "update_parameters",
    "utils",
    "value_and_grad",
    "VanillaRNN",
    "VanillaRNNState",
)


try:
    del _src  # pylint: disable=undefined-variable
except NameError:
    pass
